#include "spawn.h"

#include "..\algorithms.h"

#include "..\dllcore.cpp"
#include "..\datatypes.cpp"
#include "..\algorithms.cpp"
#include "..\player.cpp"
#include "..\commtypes.cpp"


//////// Bot list ////////

_linkedlist <botInfo> botlist;

botInfo *findBot(CALL_HANDLE handle)
{
	_listnode <botInfo> *parse = botlist.head;

	while (parse)
	{
		botInfo *item = parse->item;

		if (item->validate(handle))
			return item;

		parse = parse->next;
	}

	return 0;
}


//////// DLL "import" ////////

void botInfo::tell(BotEvent &event)
{
	if (callback && handle)
	{
		event.handle = handle;
		callback(event);
	}
}

void botInfo::sendReset()
{
	if (*arena && !isNumeric(arena))
	{	// if it's in a private arena
		sendPublic("*scorereset");
	}
}

bool botInfo::sendGreet()
{
	if (!allowZone) return false;

	countdown[1] = 120;
	allowZone = false;

	String s;
	s += "*zone Elimination is starting. Type ?go ";
	s += arena;
	sendPublic(s);

	return true;
}

void botInfo::findWinner()
{
	_listnode <Player> *parse = playerlist->head;

	Player *winner = 0;
	int total = 0;

	while (parse)
	{
		Player *p = parse->item;

		if (p->ship != SHIP_Spectator)
		{
			++total;
			winner = p;

			if (total > 1)
				return;
		}

		parse = parse->next;
	}

	if (total == 0)
	{
		sendPublic(SND_VictoryBell, "*arena No winner.");
	}
	else if (total == 1)
	{
		String s;
		s += "*arena Game over. Winner: ";
		s += winner->name;
		sendPublic(SND_VictoryBell, s);
	}

	if (getIngame() >= 2)
	{
		state = ELIM_Announcing;
		countdown[0] = 5;
	}
	else
	{
		state = ELIM_EnterGame;
		countdown[0] = 3;
	}

	sendTeam("You may enter");
	sendPublic("*lock");
}

int botInfo::getIngame()
{
	_listnode <Player> *parse = playerlist->head;

	int total = 0;

	while (parse)
	{
		Player *p = parse->item;

		if (p->ship != SHIP_Spectator)
			++total;

		parse = parse->next;
	}

	return total;
}

String botInfo::getStatus()
{
	String s;

	switch (state)
	{
	case ELIM_NotPlaying:
		{
			s += "Elimination has not been started. Send ::!help for commands";
		}
		break;
	case ELIM_Voting:
		{
			s += "We are voting on the number of deaths. (1-9)";
		}
		break;
	case ELIM_GetReady:
		{
			s += "We are playing elimination to ";
			s += deaths;
			if (deaths == 1) s += " death.";
			else			 s += " deaths.";
			s += " Game will begin in ";
			s += countdown[0];
			if (countdown[0] == 1)	s += " second";
			else				s += " seconds";
		}
		break;
	case ELIM_Playing:
		{
			s += "We are playing elimination to ";
			s += deaths;
			if (deaths == 1) s += " death. ";
			else			 s += " deaths. ";
			s += getIngame();
			s += " players remaining";
		}
		break;
	case ELIM_EnterGame:
		{
			s += "Elimination will restart when 2 players enter. Send ::!help for commands";
		}
		break;
	default:
		{
			s += "I don't know what is going on!";
		}
		break;
	};

	return s;
}

void botInfo::gotEvent(BotEvent &event)
{
	if (CONNECTION_DENIED) return;

	switch (event.code)
	{
//////// Periodic ////////
	case EVENT_Tick:
		{
			for (int i = 0; i < 4; ++i)
				--countdown[i];

			if (countdown[0] == 0)
			switch (state)
			{
			case ELIM_NotPlaying:
				{
					// do nothing
				}
				break;
			case ELIM_Announcing:
				{
					// announce game, wait 20

					for (int i = 0; i < 10; ++i)
						votes[i] = 0;

					sendGreet();
					sendPublic(SND_TrebleBeep, "*arena Elimination to how many deaths? (1-9)");

					state = ELIM_Voting;
					countdown[0] = 20;
				}
				break;
			case ELIM_Voting:
				{
					// tally votes, wait 10

					int best = 2;

					for (int i = 0; i < 10; ++i)
						if (votes[best] < votes[i]) best = i;

					deaths = best;

					String s;
					s += "*arena Playing to ";
					s += deaths;
					s += ". Game will begin in 10 seconds...";
					sendPublic(SND_TrebleBeep, s);

					state = ELIM_GetReady;
					countdown[0] = 10;
				}
				break;
			case ELIM_GetReady:
				{
					// gogogo or cancel
					if (getIngame() >= 2)
					{
						sendPublic("*lock");
						sendPublic("*scorereset");
						sendPublic(SND_Goal, "*arena Go! Go! Go!");

						state = ELIM_Playing;
					}
					else
					{
						sendPublic(SND_CantLogIn, "*arena Aborted. Elimination will restart when 2 or more players enter.");

						state = ELIM_EnterGame;
						countdown[0] = 3;
					}
				}
				break;
			case ELIM_Playing:
				{
					// do nothing
				}
				break;
			case ELIM_EnterGame:
				{
					if (getIngame() >= 2)
					{
						state = ELIM_Announcing;
						countdown[0] = 1;
					}
					else
					{
						countdown[0] = 3;
					}
				}
				break;
			};

			if (countdown[1] == 0)
			{	// 2 minutes are up
				allowZone = true;
			}
		}
		break;
//////// Arena ////////
	case EVENT_ArenaEnter:
		{
			arena = (char*)event.p[0];
			me = (Player*)event.p[1];	// if(me) {/*we are in the arena*/}
		}
		break;
	case EVENT_ArenaSettings:
		{
			settings = (arenaSettings*)event.p[0];
		}
		break;
	case EVENT_ArenaLeave:
		{
			me = 0;
			state = ELIM_NotPlaying;
		}
		break;
	case EVENT_ArenaListEntry:
		{
			char *name = (char*)event.p[0];
			bool current = *(bool*)&event.p[1];
			int population = *(int*)&event.p[2];
		}
		break;
	case EVENT_ArenaListEnd:
		{
			char *name = (char*)event.p[0];
			bool current = *(bool*)&event.p[1];
			int population = *(int*)&event.p[2];
		}
		break;
//////// Flag ////////
	case EVENT_FlagGrab:
		{
			Player *p = (Player*)event.p[0];
			Flag *f = (Flag*)event.p[1];
		}
		break;
	case EVENT_FlagDrop:
		{
			Player *p = (Player*)event.p[0];
		}
		break;
	case EVENT_FlagMove:
		{
			Flag *f = (Flag*)event.p[0];
		}
		break;
	case EVENT_FlagVictory:
		{
			int team = *(int*)&event.p[0];
			int reward = *(int*)&event.p[1];
		}
		break;
	case EVENT_FlagReward:
		{
			int team = *(int*)&event.p[0];
			int reward = *(int*)&event.p[1];
		}
		break;
//////// Timed ////////
	case EVENT_TimedGameOver:
		{
			Player *p[5];
			for (int i = 0; i < 5; i++)
				p[i] = (Player*)event.p[i];
		}
		break;
//////// Soccer ////////
	case EVENT_SoccerGoal:
		{
			int team = *(int*)&event.p[0];
			int reward = *(int*)&event.p[1];
		}
		break;
//////// File ////////
	case EVENT_File:
		{
			char *name = (char*)event.p[0];
		}
		break;
//////// Player ////////
	case EVENT_PlayerEntering:
		{
			Player *p = (Player*)event.p[0];

			if (state == ELIM_Playing)
			{
				sendPrivate(p, getStatus());
			}
		}
		break;
	case EVENT_PlayerMove:
		{
			Player *p = (Player*)event.p[0];
		}
		break;
	case EVENT_PlayerDeath:
		{
			Player *p = (Player*)event.p[0],
				   *k = (Player*)event.p[1];

			if (state == ELIM_Playing)
			{
				if (p->score.losses >= deaths)
				{
					sendPrivate(p, "*spec");
					sendPrivate(p, "*spec");
				}
			}
		}
		break;
	case EVENT_PlayerScore:
		{
			Player *p = (Player*)event.p[0];

			if (state == ELIM_Playing)
			{
				if (p->score.losses >= deaths)
				{
					sendPrivate(p, "*spec");
					sendPrivate(p, "*spec");
				}
			}
		}
		break;
	case EVENT_PlayerPrize:
		{
			Player *p = (Player*)event.p[0];
		}
		break;
	case EVENT_PlayerShip:
		{
			Player *p = (Player*)event.p[0];
		}
		break;
	case EVENT_PlayerSpec:
		{
			Player *p = (Player*)event.p[0];

			if (state == ELIM_Playing)
				findWinner();
		}
		break;
	case EVENT_PlayerTeam:
		{
			Player *p = (Player*)event.p[0];
		}
		break;
	case EVENT_PlayerLeaving:
		{
			Player *p = (Player*)event.p[0];

			if (state == ELIM_Playing)
				findWinner();
		}
		break;
//////// Turret ////////
	case EVENT_CreateTurret:
		{
			Player *turreter = (Player*)event.p[0];
			Player *turretee = (Player*)event.p[1];
		}
		break;
	case EVENT_DeleteTurret:
		{
			Player *turreter = (Player*)event.p[0];
			Player *turretee = (Player*)event.p[1];
		}
		break;
//////// Chat ////////
	case EVENT_Chat:
		{
			int type  = *(int*)&event.p[0];
			int sound = *(int*)&event.p[1];
			Player *p = (Player*)event.p[2];
			char *msg = (char*)event.p[3];

			switch (type)
			{
			case MSG_Arena:
				{
					if (state == ELIM_Playing)
					{
						if (CMPSTR(msg, "Arena UNLOCKED"))
						{
							sendPublic("*lock");
						}
					}
					else
					{
						if (CMPSTR(msg, "Arena LOCKED"))
						{
							sendPublic("*lock");
						}
					}
				}
				break;
			case MSG_PublicMacro:
			case MSG_Public:
			case MSG_Team:
			case MSG_TeamPrivate:
			case MSG_Private:			if (!p) break;
				{
					if (isNumeric(msg))
					{
						int vote = getInteger(msg, 10);
						if (vote < 1 || vote > 9) break;

						++votes[vote];
					}
				}
				break;
			case MSG_PlayerWarning:		if (!p) break;
				{
				}
				break;
			case MSG_RemotePrivate:
				{
				}
				break;
			case MSG_ServerError:
				{
				}
				break;
			case MSG_Channel:
				{
				}
				break;
			};
		}
		break;
	case EVENT_LocalCommand:
		{
			Player *p = (Player*)event.p[0];
			Command *c = (Command*)event.p[1];

			gotCommand(p, c);
		}
		break;
	case EVENT_LocalHelp:
		{
			Player *p = (Player*)event.p[0];
			Command *c = (Command*)event.p[1];

			gotHelp(p, c);
		}
		break;
	case EVENT_RemoteCommand:
		{
			char *p = (char*)event.p[0];
			Command *c = (Command*)event.p[1];
			Operator_Level o = *(Operator_Level*)&event.p[2];

			gotRemote(p, c, o);
		}
		break;
	case EVENT_RemoteHelp:
		{
			char *p = (char*)event.p[0];
			Command *c = (Command*)event.p[1];
			Operator_Level o = *(Operator_Level*)&event.p[2];

			gotRemoteHelp(p, c, o);
		}
		break;
//////// Containment ////////
	case EVENT_Init:
		{
			int major = HIWORD(*(int*)&event.p[0]);
			int minor = LOWORD(*(int*)&event.p[0]);
			callback = (CALL_COMMAND)event.p[1];
			playerlist = (CALL_PLIST)event.p[2];
			flaglist = (CALL_FLIST)event.p[3];

			if (major != CORE_MAJOR_VERSION)
			{
				tell(makeEcho("DLL plugin cannot connect.  Major DLL version mismatch!"));

				CONNECTION_DENIED = true;

				return;
			}
			else if (minor != CORE_MINOR_VERSION)
				tell(makeEcho("DLL plugin connected.  Minor DLL version mismatch!"));
			else
				tell(makeEcho("DLL plugin connected."));
		}
		break;
	case EVENT_Term:
		{
			tell(makeEcho("DLL plugin disconnected."));
		}
		break;
	default:
		{
			String s;
			s += "Unknown message from bot (";
			s += event.code;
			s += ")  Version mismatch?";
			tell(makeEcho(s));
		}
		break;
	};
}


//////// DLL export ////////

_declspec(dllexport) void __stdcall talk(BotEvent &event)
{
	botInfo *bot;

	bot = findBot(event.handle);

	if (!bot)
	{
		bot = new botInfo(event.handle);
		botlist.append(bot);
	}

	if (bot) bot->gotEvent(event);
}


//////// Chatter ////////

void botInfo::sendPrivate(Player *player, BYTE snd, char *msg)
{
	tell(makeSay(MSG_Private, snd, player->ident, msg));
}

void botInfo::sendPrivate(Player *player, char *msg)
{
	tell(makeSay(MSG_Private, 0, player->ident, msg));
}

void botInfo::sendTeam(char *msg)
{
	tell(makeSay(MSG_Team, 0, 0, msg));
}

void botInfo::sendTeam(BYTE snd, char *msg)
{
	tell(makeSay(MSG_Team, snd, 0, msg));
}

void botInfo::sendTeamPrivate(Uint16 team, char *msg)
{
	tell(makeSay(MSG_TeamPrivate, 0, team, msg));
}

void botInfo::sendTeamPrivate(Uint16 team, BYTE snd, char *msg)
{
	tell(makeSay(MSG_TeamPrivate, snd, team, msg));
}

void botInfo::sendPublic(char *msg)
{
	tell(makeSay(MSG_Public, 0, 0, msg));
}

void botInfo::sendPublic(BYTE snd, char *msg)
{
	tell(makeSay(MSG_Public, snd, 0, msg));
}

void botInfo::sendPublicMacro(char *msg)
{
	tell(makeSay(MSG_PublicMacro, 0, 0, msg));
}

void botInfo::sendPublicMacro(BYTE snd, char *msg)
{
	tell(makeSay(MSG_PublicMacro, snd, 0, msg));
}

void botInfo::sendChannel(char *msg)
{
	tell(makeSay(MSG_Channel, 0, 0, msg));
}

void botInfo::sendRemotePrivate(char *msg)
{
	tell(makeSay(MSG_RemotePrivate, 0, 0, msg));
}

void botInfo::sendRemotePrivate(char *name, char *msg)
{
	String s;
	s += ":";
	s += name;
	s += ":";
	s += msg;

	sendRemotePrivate(s);
}
